@@ -47,6 +47,7 @@
 #include "gmarkup.h"
 #include "gmem.h"
 #include "gmessages.h"
+#include "gshell.h"
 #include "gslice.h"
 #include "gstdio.h"
 #include "gstring.h"
@@ -237,11 +238,8 @@ bookmark_app_info_free (BookmarkAppInfo *app_info)
   if (!app_info)
     return;
   
-  if (app_info->name)
-    g_free (app_info->name);
-  
-  if (app_info->exec)
-    g_free (app_info->exec);
+  g_free (app_info->name);
+  g_free (app_info->exec);
   
   g_slice_free (BookmarkAppInfo, app_info);
 }
@@ -311,8 +309,7 @@ bookmark_metadata_free (BookmarkMetadata *metadata)
   if (!metadata)
     return;
   
-  if (metadata->mime_type)
-    g_free (metadata->mime_type);
+  g_free (metadata->mime_type);
     
   if (metadata->groups)
     {
@@ -332,11 +329,8 @@ bookmark_metadata_free (BookmarkMetadata *metadata)
       g_hash_table_destroy (metadata->apps_by_name);
     }
   
-  if (metadata->icon_href)
-    g_free (metadata->icon_href);
-  
-  if (metadata->icon_mime)
-    g_free (metadata->icon_mime);
+  g_free (metadata->icon_href);
+  g_free (metadata->icon_mime);
   
   g_slice_free (BookmarkMetadata, metadata);
 }
@@ -491,14 +485,9 @@ bookmark_item_free (BookmarkItem *item)
   if (!item)
     return;
 
-  if (item->uri)
-    g_free (item->uri);
-  
-  if (item->title)
-    g_free (item->title);
-  
-  if (item->description)
-    g_free (item->description);
+  g_free (item->uri);
+  g_free (item->title);
+  g_free (item->description);
   
   if (item->metadata)
     bookmark_metadata_free (item->metadata);
@@ -1789,7 +1778,7 @@ find_file_in_data_dirs (const gchar   *file,
     {
       g_set_error (error, G_BOOKMARK_FILE_ERROR,
                    G_BOOKMARK_FILE_ERROR_FILE_NOT_FOUND,
-                   _("No valid bookmark file was be found in data dirs"));
+                   _("No valid bookmark file found in data dirs"));
       
       return NULL;
     }
@@ -2319,8 +2308,7 @@ g_bookmark_file_set_mime_type (GBookmarkFile *bookmark,
   if (!item->metadata)
     item->metadata = bookmark_metadata_new ();
   
-  if (item->metadata->mime_type != NULL)
-    g_free (item->metadata->mime_type);
+  g_free (item->metadata->mime_type);
   
   item->metadata->mime_type = g_strdup (mime_type);
   item->modified = time (NULL);
@@ -2842,7 +2830,8 @@ g_bookmark_file_remove_group (GBookmarkFile  *bookmark,
       if (strcmp (l->data, group) == 0)
         {
           item->metadata->groups = g_list_remove_link (item->metadata->groups, l);
-          g_list_free_1 (l);
+          g_free (l->data);
+	  g_list_free_1 (l);
           
           item->modified = time (NULL);          
           
@@ -2926,6 +2915,8 @@ g_bookmark_file_set_groups (GBookmarkFile  *bookmark,
  *
  * Return value: a newly allocated %NULL-terminated array of group names.
  *   Use g_strfreev() to free it.
+ *
+ * Since: 2.12
  */
 gchar **
 g_bookmark_file_get_groups (GBookmarkFile  *bookmark,
@@ -3248,7 +3239,8 @@ g_bookmark_file_set_app_info (GBookmarkFile  *bookmark,
     {
       item->metadata->applications = g_list_remove (item->metadata->applications, ai);
       g_hash_table_remove (item->metadata->apps_by_name, ai->name);
-          
+      bookmark_app_info_free (ai);
+
       item->modified = time (NULL);
           
       return TRUE;
@@ -3266,7 +3258,7 @@ g_bookmark_file_set_app_info (GBookmarkFile  *bookmark,
   if (exec && exec[0] != '\0')
     {
       g_free (ai->exec);
-      ai->exec = g_strdup (exec);
+      ai->exec = g_shell_quote (exec);
     }
   
   item->modified = time (NULL);
@@ -3296,14 +3288,24 @@ expand_exec_line (const gchar *exec_fmt,
        {
        case '\0':
 	 goto out;
+       case 'U':
        case 'u':
          g_string_append (exec, uri);
          break;
+       case 'F':
        case 'f':
          {
 	   gchar *file = g_filename_from_uri (uri, NULL, NULL);
-	   g_string_append (exec, file);
-	   g_free (file);
+           if (file)
+             {
+	       g_string_append (exec, file);
+	       g_free (file);
+             }
+           else
+             {
+               g_string_free (exec, TRUE);
+               return NULL;
+             }
          }
          break;
        case '%':
@@ -3337,7 +3339,9 @@ expand_exec_line (const gchar *exec_fmt,
  * @error is set to #G_BOOKMARK_FILE_ERROR_URI_NOT_FOUND.  In the
  * event that no application with name @app_name has registered a bookmark
  * for @uri,  %FALSE is returned and error is set to
- * #G_BOOKMARK_FILE_ERROR_APP_NOT_REGISTERED.
+ * #G_BOOKMARK_FILE_ERROR_APP_NOT_REGISTERED. In the event that unquoting
+ * the command line fails, an error of the #G_SHELL_ERROR domain is
+ * set and %FALSE is returned.
  *
  * Return value: %TRUE on success.
  *
@@ -3381,8 +3385,32 @@ g_bookmark_file_get_app_info (GBookmarkFile  *bookmark,
     }
   
   if (exec)
-    *exec = expand_exec_line (ai->exec, uri);
-  
+    {
+      GError *unquote_error = NULL;
+      gchar *command_line;
+      
+      command_line = g_shell_unquote (ai->exec, &unquote_error);
+      if (unquote_error)
+        {
+          g_propagate_error (error, unquote_error);
+          return FALSE;
+        }
+
+      *exec = expand_exec_line (command_line, uri);
+      if (!*exec)
+        {
+          g_set_error (error, G_BOOKMARK_FILE_ERROR,
+		       G_BOOKMARK_FILE_ERROR_INVALID_URI,
+		       _("Failed to expand exec line '%s' with URI '%s'"),
+		     ai->exec, uri);
+          g_free (command_line);
+
+          return FALSE;
+        }
+      else
+        g_free (command_line);
+    } 
+
   if (count)
     *count = ai->count;
   
