@@ -19,20 +19,25 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <config.h>
+#include "config.h"
 
-#include "gregex.h"
-
-#include <glib.h>
-#include <glib/gi18n.h>
 #include <string.h>
 
+#include "glib.h"
+#include "glibintl.h"
+#include "gregex.h"
+
 #ifdef USE_SYSTEM_PCRE
 #include <pcre.h>
 #else
 #include "pcre/pcre.h"
 #endif
 
+/* PCRE 7.3 does not contain the definition of PCRE_ERROR_NULLWSLIMIT */
+#ifndef PCRE_ERROR_NULLWSLIMIT
+#define PCRE_ERROR_NULLWSLIMIT (-22)
+#endif
+
 #include "galias.h"
 
 /* Mask of all the possible values for GRegexCompileFlags. */
@@ -87,7 +92,7 @@ struct _GMatchInfo
 
 struct _GRegex
 {
-  volatile guint ref_count;	/* the ref count for the immutable part */
+  volatile gint ref_count;	/* the ref count for the immutable part */
   gchar *pattern;		/* the pattern */
   pcre *pcre_re;		/* compiled form of the pattern */
   GRegexCompileFlags compile_opts;	/* options used at compile time on the pattern */
@@ -349,7 +354,7 @@ g_match_info_next (GMatchInfo  *match_info,
  * g_match_info_matches:
  * @match_info: a #GMatchInfo structure
  *
- * Returns wether the previous match operation succeeded.
+ * Returns whether the previous match operation succeeded.
  * 
  * Returns: %TRUE if the previous match operation succeeded, 
  *   %FALSE otherwise
@@ -625,20 +630,8 @@ get_matched_substring_number (const GMatchInfo *match_info,
   gchar *first, *last;
   guchar *entry;
 
-  /*
-   * FIXME: (?J) may be used inside the pattern as the equivalent of
-   * DUPNAMES compile option. In this case we can't know about it,
-   * and pcre doesn't tell us about it either, it uses private flag
-   * PCRE_JCHANGED for this. So we have to always search string
-   * table, unlike pcre which uses pcre_get_stringnumber() shortcut
-   * when possible. It shouldn't be actually bad since
-   * pcre_get_stringtable_entries() uses binary search; still would 
-   * be better to fix it, to be not worse than pcre.
-   */
-#if 0
-  if ((match_info->regex->compile_opts & G_REGEX_DUPNAMES) == 0)
+  if (!(match_info->regex->compile_opts & G_REGEX_DUPNAMES))
     return pcre_get_stringnumber (match_info->regex->pcre_re, name);
-#endif
 
   /* This code is copied from pcre_get.c: get_first_set() */
   entrysize = pcre_get_stringtable_entries (match_info->regex->pcre_re, 
@@ -768,7 +761,7 @@ g_match_info_fetch_all (const GMatchInfo *match_info)
   gchar **result;
   gint i;
 
-  g_return_val_if_fail (match_info != NULL, FALSE);
+  g_return_val_if_fail (match_info != NULL, NULL);
 
   if (match_info->matches < 0)
     return NULL;
@@ -809,7 +802,7 @@ GRegex *
 g_regex_ref (GRegex *regex)
 {
   g_return_val_if_fail (regex != NULL, NULL);
-  g_atomic_int_inc ((gint*) &regex->ref_count);
+  g_atomic_int_inc (&regex->ref_count);
   return regex;
 }
 
@@ -827,7 +820,7 @@ g_regex_unref (GRegex *regex)
 {
   g_return_if_fail (regex != NULL);
 
-  if (g_atomic_int_exchange_and_add ((gint *) &regex->ref_count, -1) - 1 == 0)
+  if (g_atomic_int_exchange_and_add (&regex->ref_count, -1) - 1 == 0)
     {
       g_free (regex->pattern);
       if (regex->pcre_re != NULL)
@@ -865,7 +858,7 @@ g_regex_new (const gchar         *pattern,
   gint erroffset;
   gboolean optimize = FALSE;
   static gboolean initialized = FALSE;
-  
+
   g_return_val_if_fail (pattern != NULL, NULL);
   g_return_val_if_fail (error == NULL || *error == NULL, NULL);
   g_return_val_if_fail ((compile_options & ~G_REGEX_COMPILE_MASK) == 0, NULL);
@@ -941,6 +934,19 @@ g_regex_new (const gchar         *pattern,
       return NULL;
     }
 
+  /* For options set at the beginning of the pattern, pcre puts them into
+   * compile options, e.g. "(?i)foo" will make the pcre structure store
+   * PCRE_CASELESS even though it wasn't explicitly given for compilation. */
+  pcre_fullinfo (re, NULL, PCRE_INFO_OPTIONS, &compile_options);
+
+  if (!(compile_options & G_REGEX_DUPNAMES))
+    {
+      gboolean jchanged = FALSE;
+      pcre_fullinfo (re, NULL, PCRE_INFO_JCHANGED, &jchanged);
+      if (jchanged)
+	compile_options |= G_REGEX_DUPNAMES;
+    }
+
   regex = g_new0 (GRegex, 1);
   regex->ref_count = 1;
   regex->pattern = g_strdup (pattern);
@@ -1670,7 +1676,7 @@ g_regex_split_full (const GRegex      *regex,
   i = 0;
   for (last = g_list_last (list); last; last = g_list_previous (last))
     string_list[i++] = last->data;
-  string_list[i] = 0;
+  string_list[i] = NULL;
   g_list_free (list);
 
   return string_list;
